-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt
-- Author: Funky7Monkey

local C = {}
C.__index = C

local function rotateEuler(x, y, z, q)
  q = q or quat()
  q = quatFromEuler(0, z, 0) * q
  q = quatFromEuler(0, 0, x) * q
  q = quatFromEuler(y, 0, 0) * q
  return q
end

function C:init()
  self.veloSmoother = newExponentialSmoothing(20, 1)
  self.lastDataPos = vec3()
  self.camLastPos = vec3(0,0,40)
  self.camLastVel = vec3(0,0,0)
  self.camLastDist = vec3(0,0,0)
  self.fov = self.fov or 25
  self.rot = vec3()
  self.target = 0
  self.camReseted = 2
  self.dist = 60
  self.dump = 0

  --- this marks it as base mode. It needs to set position, rotation and fov correctly
  self.baseMode = true
  -- add it to the default camera modes?
  self.register = true
end

function C:reset()
  self.camReseted = 2
end

function C:reload()
  self.camReseted = 1
end

function C:lookback()
  if self.target == 0 then
    self.target = 1
    ui_message("Helicam position: Hold")
  else
    self.target = 0
    ui_message("Helicam position: Normal")
  end
end

function C:update(data)
  if self.dump == 0 then dump(data) self.dump = 1 end
  local ref  = vec3(data.veh:getNodePosition(self.refNodes.ref))
  local back = vec3(data.veh:getNodePosition(self.refNodes.back))

  local velo = self.veloSmoother:get((self.lastDataPos - data.pos):length())

  local dtfactor = data.dt * 300
  local dx = (MoveManager.yawLeftSpeed - MoveManager.yawRightSpeed) * (dtfactor * -1)
  local dy = (MoveManager.pitchDownSpeed - MoveManager.pitchUpSpeed) * dtfactor
  local dz = (BeamEngine.zoomInSpeed - BeamEngine.zoomOutSpeed) * dtfactor * 10
  self.rot = self.rot + vec3(dx, dy, 0)

  if vec3(dx,dy,0) ~= vec3(0,0,0) then if self.target == 0 then
    self.target = 1
    ui_message("Helicam position: Hold")
  end end

  local dir = (ref - back):normalized()
  local targetPos = data.pos + vec3(dir.x, dir.y, 0) * math.min(25, 10 * velo) * (self.camLastDist:length()/100)
  local camDist = (targetPos - self.camLastPos)
  local camLastVel = self.camLastVel or vec3(0,0,0)
  local camVel = vec3(0,0,0)
  local camPos = vec3()

  if self.camReseted == 2 then
    camPos = targetPos - vec3(dir.x, dir.y, 0) * 30
    camPos['z'] = targetPos['z'] + 20
    self.rot = camPos - targetPos
    self.camReseted = 0
  elseif self.camReseted == 1 then
    if camDist:length() > self.dist then
      camPos = targetPos - vec3(dir.x, dir.y, 0) * 30
      camPos['z'] = targetPos['z'] + 20
      self.rot = camPos - targetPos
      self.camReseted = 0
    end
  elseif self.target == 1 then
    camPos = self.camLastPos + vec3(dx,-dy,0)
  else
    if camLastVel['x'] == 0 then
      if camDist:length() > self.dist then
        camVel['x'] = self.camLastVel['x'] + (camDist['x']/6000)
      else
        camVel['x'] = 0
      end
    else
      if camDist:length() > self.dist then
        camVel['x'] = self.camLastVel['x'] + (camDist['x']/6000)
      else
        if math.floor(self.camLastVel['x']*5+0.5)/5 ~= 0 then
          camVel['x'] = self.camLastVel['x'] - (camDist['x']/2000)
        else
          camVel['x'] = 0
        end
        if self.camLastDist:length() < camDist:length() then
          if self.camLastDist ~= vec3(0,0,0) then
            camVel['x'] = camVel['x'] / 1.25
          end
        end
      end
    end
    if camLastVel['y'] == 0 then
      if camDist:length() > self.dist then
        camVel['y'] = self.camLastVel['y'] + (camDist['y']/6000)
      else
        camVel['y'] = 0
      end
    else
      if camDist:length() > self.dist then
        camVel['y'] = self.camLastVel['y'] + (camDist['y']/6000)
      else
        if math.floor(self.camLastVel['y']*5+0.5)/5 ~= 0 then
          camVel['y'] = self.camLastVel['y'] - (camDist['y']/2000)
        else
          camVel['y'] = 0
        end
        if self.camLastDist:length() < camDist:length() then
          if self.camLastDist ~= vec3(0,0,0) then
            camVel['y'] = camVel['y'] / 1.25
          end
        end
      end
    end
    camPos = self.camLastPos + camVel
  end
  camPos['z'] = camPos['z'] + dz
  local qdir = quatFromDir((targetPos - camPos):normalized())

  self.fov = 20 + math.min(30, 15 * velo)
  self.lastDataPos = data.pos
  self.camLastVel = camVel
  self.camLastDist = camDist
  self.camLastPos = camPos

  -- set the data, this needs to happen
  data.res.pos = camPos -- required, vec3()
  data.res.rot = qdir -- required, quat()
  data.res.fov = self.fov -- required
  data.res.targetPos = targetPos -- this is optional
  return true
end

-- DO NOT CHANGE CLASS IMPLEMENTATION BELOW

return function(...)
  local o = ... or {}
  setmetatable(o, C)
  o:init()
  return o
end
